# Generated by Django 3.0.5 on 2020-04-25 10:24

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid
import swapper


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='PlanList',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.TextField(blank=True, help_text='title to display on the subscription plan list page',
                                           null=True)),
                ('slug',
                 models.SlugField(blank=True, help_text='slug to reference the subscription plan list', max_length=128,
                                  null=True, unique=True)),
                ('subtitle',
                 models.TextField(blank=True, help_text='subtitle to display on the subscription plan list page',
                                  null=True)),
                ('header',
                 models.TextField(blank=True, help_text='header text to display on the subscription plan list page',
                                  null=True)),
                ('footer',
                 models.TextField(blank=True, help_text='header text to display on the subscription plan list page',
                                  null=True)),
                ('active', models.BooleanField(default=True, help_text='whether this plan list is active or not.')),
            ],
        ),
        migrations.CreateModel(
            name='PlanTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tag', models.CharField(help_text='the tag name', max_length=64, unique=True)),
            ],
            options={
                'ordering': ('tag',),
            },
        ),
        migrations.CreateModel(
            name='SubscriptionPlan',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False,
                                        verbose_name='ID')),
                ('plan_name', models.CharField(help_text='the name of the subscription plan', max_length=128)),
                ('slug',
                 models.SlugField(blank=True, help_text='slug to reference the subscription plan', max_length=128,
                                  null=True, unique=True)),
                ('plan_description',
                 models.CharField(blank=True, help_text='a description of the subscription plan', max_length=512,
                                  null=True)),
                ('grace_period', models.PositiveIntegerField(default=0,
                                                             help_text='how many days after the subscription ends before the subscription expires')),
                ('group', models.ForeignKey(blank=True, help_text='the Django auth group for this plan', null=True,
                                            on_delete=django.db.models.deletion.SET_NULL, related_name='plans',
                                            to='auth.Group')),
                ('tags', models.ManyToManyField(blank=True, help_text='any tags associated with this plan',
                                                related_name='plans', to='subscriptions_api.PlanTag')),
            ],
            options={
                'ordering': ('plan_name',),
                'permissions': (('subscriptions', 'Can interact with subscription details'),),
            },
        ),
        migrations.CreateModel(
            name='PlanListDetail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('html_content', models.TextField(blank=True, help_text='HTML content to display for plan', null=True)),
                ('subscribe_button_text', models.CharField(blank=True, default='Subscribe', max_length=128, null=True)),
                ('order', models.PositiveIntegerField(default=1,
                                                      help_text='Order to display plan in (lower numbers displayed first)')),
                ('plan',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plan_list_details',
                                   to='subscriptions_api.SubscriptionPlan')),
                ('plan_list',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plan_list_details',
                                   to='subscriptions_api.PlanList')),
            ],
        ),
        migrations.CreateModel(
            name='PlanCost',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False,
                                        verbose_name='ID')),
                ('slug', models.SlugField(blank=True, help_text='slug to reference these cost details', max_length=128,
                                          null=True, unique=True)),
                ('recurrence_period', models.PositiveSmallIntegerField(default=1,
                                                                       help_text='how often the plan is billed (per recurrence unit)',
                                                                       validators=[
                                                                           django.core.validators.MinValueValidator(
                                                                               1)])),
                ('recurrence_unit', models.CharField(
                    choices=[('0', 'once'), ('1', 'second'), ('2', 'minute'), ('3', 'hour'), ('4', 'day'),
                             ('5', 'week'), ('6', 'month'), ('7', 'year')], default='6', max_length=1)),
                ('cost',
                 models.DecimalField(blank=True, decimal_places=2, help_text='the cost per recurrence of the plan',
                                     max_digits=19, null=True)),
                ('plan', models.ForeignKey(help_text='the subscription plan for these cost details',
                                           on_delete=django.db.models.deletion.CASCADE, related_name='costs',
                                           to='subscriptions_api.SubscriptionPlan')),
            ],
            options={
                'ordering': ('recurrence_unit', 'recurrence_period', 'cost'),
            },
        ),
        migrations.CreateModel(
            name='UserSubscription',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False,
                                        verbose_name='ID')),
                ('date_billing_start',
                 models.DateTimeField(blank=True, help_text='the date to start billing this subscription', null=True,
                                      verbose_name='billing start date')),
                ('date_billing_end',
                 models.DateTimeField(blank=True, help_text='the date to finish billing this subscription', null=True,
                                      verbose_name='billing start end')),
                ('date_billing_last',
                 models.DateTimeField(blank=True, help_text='the last date this plan was billed', null=True,
                                      verbose_name='last billing date')),
                ('date_billing_next',
                 models.DateTimeField(blank=True, help_text='the next date billing is due', null=True,
                                      verbose_name='next start date')),
                ('active', models.BooleanField(default=True, help_text='whether this subscription is active or not')),
                ('due', models.BooleanField(default=False, help_text='whether this subscription is due or not')),
                ('cancelled',
                 models.BooleanField(default=False, help_text='whether this subscription is cancelled or not')),
                ('plan_cost',
                 models.ForeignKey(help_text='the plan costs and billing frequency for this user', null=True,
                                   on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions',
                                   to='subscriptions_api.PlanCost')),
                ('user', models.ForeignKey(help_text='the user this subscription applies to', null=True,
                                           on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions',
                                           to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'swappable': swapper.swappable_setting('subscriptions_api', 'UserSubscription'),
            },
        ),
        migrations.CreateModel(
            name='SubscriptionTransaction',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False,
                                        verbose_name='ID')),
                ('date_transaction', models.DateTimeField(help_text='the datetime the transaction was billed',
                                                          verbose_name='transaction date')),
                ('amount',
                 models.DecimalField(blank=True, decimal_places=2, help_text='how much was billed for the user',
                                     max_digits=19, null=True)),
                ('paid', models.BooleanField(default=False, help_text='Mark transaction has paid')),
                ('subscription', models.ForeignKey(help_text='the Subscription that were billed', null=True,
                                                   on_delete=django.db.models.deletion.SET_NULL,
                                                   related_name='transactions',
                                                   to=swapper.get_model_name('subscriptions_api', 'UserSubscription'))),
                ('user', models.ForeignKey(help_text='the user that this subscription was billed for', null=True,
                                           on_delete=django.db.models.deletion.SET_NULL,
                                           related_name='subscription_transactions', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'swappable': swapper.swappable_setting('subscriptions_api', 'SubscriptionTransaction'),
            },
        ),
    ]



from rest_framework.permissions import BasePermission, SAFE_METHODS


class IsAdminOrReadOnly(BasePermission):
    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        else:
            return request.user.is_staff



import json
import swapper
from rest_framework import serializers
from subscriptions_api.plans import get_planlist_features
from subscriptions_api import models

UserSubscriptionModel = swapper.load_model('subscriptions_api', 'UserSubscription')
SubscriptionTransactionModel = swapper.load_model('subscriptions_api', 'SubscriptionTransaction')


class PlanTagSerializer(serializers.ModelSerializer):
    """Serializer for PlanTag model"""

    class Meta:
        model = models.PlanTag
        fields = '__all__'


class PlanCostSerializer(serializers.ModelSerializer):
    """PlanCost model serializer with property fields  exposed as serializer method fields"""
    recurrent_unit_text = serializers.SerializerMethodField()
    billing_frequency_text = serializers.SerializerMethodField()

    def get_recurrent_unit_text(self, obj):
        return obj.display_recurrent_unit_text

    def get_billing_frequency_text(self, obj):
        return obj.display_billing_frequency_text

    class Meta:
        model = models.PlanCost
        fields = '__all__'


class SubscriptionPlanSerializer(serializers.ModelSerializer):
    """Serializer for SubscriptionPlan model and Tags can be created directly on it"""

    tags = PlanTagSerializer(many=True, read_only=True)
    tags_str = serializers.SerializerMethodField()
    features = serializers.SerializerMethodField()
    costs = PlanCostSerializer(many=True, read_only=True)

    def get_tags_str(self, obj):
        return obj.display_tags()

    def get_features(self, obj):
        if obj.features:
            return json.loads(obj.features)

    class Meta:
        model = models.SubscriptionPlan
        fields = '__all__'


class PlanListDetailSerializer(serializers.ModelSerializer):
    """PlanListDetail serializer"""
    plan = SubscriptionPlanSerializer()

    class Meta:
        model = models.PlanListDetail
        fields = '__all__'

    def to_internal_value(self, data):
        self.fields['plan'] = serializers.PrimaryKeyRelatedField(queryset=models.SubscriptionPlan.objects.all())
        return super(PlanListDetailSerializer, self).to_internal_value(data)

    def to_representation(self, obj):
        self.fields['plan'] = SubscriptionPlanSerializer()
        return super(PlanListDetailSerializer, self).to_representation(obj)


class PlanListSerializer(serializers.ModelSerializer):
    """PlanList serializer"""
    plan_list_details = PlanListDetailSerializer(many=True, read_only=True)
    features = serializers.SerializerMethodField()

    class Meta:
        model = models.PlanList
        fields = '__all__'

    def get_features(self, obj):
        label = get_planlist_features(obj.features_ref)
        return label


class SubscriptionTransactionSerializer(serializers.ModelSerializer):
    """SubscriptionTransaction serializer"""

    class Meta:
        model = SubscriptionTransactionModel
        fields = '__all__'


class UserSubscriptionSerializer(serializers.ModelSerializer):
    """User subscription model serializer"""
    transactions = SubscriptionTransactionSerializer(many=True, read_only=True)
    description = serializers.SerializerMethodField()

    class Meta:
        model = UserSubscriptionModel
        fields = '__all__'

    def get_description(self, obj):
        return obj.description



import swapper
from rest_framework import viewsets
from subscriptions_api import serializers, models
from .permissions import IsAdminOrReadOnly

UserSubscriptionModel = swapper.load_model('subscriptions_api', 'UserSubscription')
SubscriptionTransactionModel = swapper.load_model('subscriptions_api', 'SubscriptionTransaction')


class PlanTagViewSet(viewsets.ModelViewSet):
    queryset = models.PlanTag.objects.all()
    serializer_class = serializers.PlanTagSerializer
    permission_classes = (IsAdminOrReadOnly,)


class SubscriptionPlanViewSet(viewsets.ModelViewSet):
    queryset = models.SubscriptionPlan.objects.all()
    serializer_class = serializers.SubscriptionPlanSerializer
    permission_classes = (IsAdminOrReadOnly,)


class PlanCostViewSet(viewsets.ModelViewSet):
    queryset = models.PlanCost.objects.all()
    serializer_class = serializers.PlanCostSerializer
    permission_classes = (IsAdminOrReadOnly,)


class UserSubscriptionViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.UserSubscriptionSerializer
    permission_classes = (IsAdminOrReadOnly,)

    def get_queryset(self):
        if self.request.user.is_staff:
            return UserSubscriptionModel.objects.all()
        return UserSubscriptionModel.objects.filter(user=self.request.user)


class SubscriptionTransactionViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.SubscriptionTransactionSerializer
    permission_classes = (IsAdminOrReadOnly,)

    def get_queryset(self):
        if self.request.user.is_staff:
            return SubscriptionTransactionModel.objects.all()
        return SubscriptionTransactionModel.objects.filter(user=self.request.user)


class PlanListViewSet(viewsets.ModelViewSet):
    queryset = models.PlanList.objects.all()
    serializer_class = serializers.PlanListSerializer
    permission_classes = (IsAdminOrReadOnly,)


class PlanListDetailViewSet(viewsets.ModelViewSet):
    queryset = models.PlanListDetail.objects.all()
    serializer_class = serializers.PlanListDetailSerializer
    permission_classes = (IsAdminOrReadOnly,)



from django.core.mail import EmailMultiAlternatives


class EmailNotification:
    """A simple class to send email notification from subscription"""

    def __init__(self, subscription, notification, **kwargs):
        self.subscription = subscription
        self.kwargs = kwargs
        self.notification = notification
        self.msg = EmailMultiAlternatives(
            subject=self.get_subject(),
            body=self.get_body(),
            from_email=self.get_from_email(),
            to=self.get_to_email(),
        )

    def get_subject(self):
        """Override and return email subject"""
        return ""

    def get_body(self):
        """Override and return email body"""
        return ""

    def get_from_email(self):
        """Override and return from email default to email from settings"""
        return None

    def get_to_email(self):
        """Override and return to email returns email of user linked to subscription"""
        return [self.subscription.user.email]

    def get_attachment_file(self):
        """Return list of attachment file paths"""
        return None

    def get_attachment(self):
        """Return tuple containing filename, filecontent and mime type"""
        return None

    def get_html_content(self):
        """Return html content to send html version of email"""
        return None

    def attach_file(self):
        files = self.get_attachment_file()
        attachment = self.get_attachment()
        if files:
            if not isinstance(files, list):
                files = [files]
            for file in files:
                self.msg.attach_file(file)
        if isinstance(attachment, tuple):
            self.msg.attach(*attachment)

    def attach_html_content(self):
        html = self.get_html_content()
        if html:
            self.msg.attach_alternative(html, "text/html")

    def extra_process(self):

        self.attach_html_content()
        self.attach_file()

    def fail_silently(self):
        return True

    def send(self):
        self.extra_process()
        self.msg.send(fail_silently=self.fail_silently())



from django.contrib import admin
from swapper import load_model
from subscriptions_api.models import PlanList, PlanListDetail, PlanTag, PlanCost, SubscriptionPlan

UserSubscription = load_model('subscriptions_api', 'UserSubscription')
SubscriptionTransaction = load_model('subscriptions_api', 'SubscriptionTransaction')

admin.site.register(PlanTag)
admin.site.register(PlanCost)
admin.site.register(SubscriptionPlan)
admin.site.register(PlanList)
admin.site.register(PlanListDetail)
admin.site.register(UserSubscription)
admin.site.register(SubscriptionTransaction)



# Generated by Django 3.0.5 on 2020-09-08 14:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions_api', '0003_subscriptionplan_feature_ref'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscriptionplan',
            name='features',
            field=models.TextField(blank=True, help_text='list of allowed features for this plan', null=True),
        ),
        migrations.AddField(
            model_name='subscriptionplan',
            name='trial_period',
            field=models.PositiveIntegerField(default=0, help_text='how many days to give free before the subscription subscription begins'),
        ),
    ]



from rest_framework import routers
from .views import PlanTagViewSet, PlanCostViewSet, PlanListDetailViewSet, \
    PlanListViewSet, SubscriptionPlanViewSet, SubscriptionTransactionViewSet, \
    UserSubscriptionViewSet

app_name = 'subscriptions_api'

router = routers.SimpleRouter()
router.register('plan-tags', PlanTagViewSet, basename='plan-tags')
router.register('plan-costs', PlanCostViewSet, basename='plan-costs')
router.register('planlist-details', PlanListDetailViewSet, basename='planlist-details')
router.register('planlist', PlanListViewSet, basename='planlist')
router.register('subscription-plans', SubscriptionPlanViewSet, basename='subscription-plans')
router.register('subscription-transactions', SubscriptionTransactionViewSet, basename='subscription-transactions')
router.register('user-subscriptions', UserSubscriptionViewSet, basename='user-subscriptions')

urlpatterns = router.urls



import swapper
from datetime import timedelta
import importlib
from uuid import uuid4
from django.db import models
from django.contrib.auth import get_user_model
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from subscriptions_api.app_settings import SETTINGS

SubscriptionTransactionModel = swapper.get_model_name('subscriptions_api', 'SubscriptionTransaction')
UserSubscriptionModel = swapper.get_model_name('subscriptions_api', 'UserSubscription')


class BaseUserSubscription(models.Model):
    """Details of a user's specific subscription."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    user = models.ForeignKey(
        get_user_model(),
        help_text=_('the user this subscription applies to'),
        null=True,
        on_delete=models.CASCADE,
        related_name='subscriptions',
    )
    plan_cost = models.ForeignKey(
        'subscriptions_api.PlanCost',
        help_text=_('the plan costs and billing frequency for this user'),
        null=True,
        on_delete=models.CASCADE,
        related_name='subscriptions'
    )
    date_billing_start = models.DateTimeField(
        blank=True,
        help_text=_('the date to start billing this subscription'),
        null=True,
        verbose_name='billing start date',
    )
    date_billing_end = models.DateTimeField(
        blank=True,
        help_text=_('the date to finish billing this subscription'),
        null=True,
        verbose_name='billing start end',
    )
    date_billing_last = models.DateTimeField(
        blank=True,
        help_text=_('the last date this plan was billed'),
        null=True,
        verbose_name='last billing date',
    )
    date_billing_next = models.DateTimeField(
        blank=True,
        help_text=_('the next date billing is due'),
        null=True,
        verbose_name='next start date',
    )
    active = models.BooleanField(
        default=True,
        help_text=_('whether this subscription is active or not'),
    )
    due = models.BooleanField(
        default=False,
        help_text=_('whether this subscription is due or not'),
    )
    cancelled = models.BooleanField(
        default=False,
        help_text=_('whether this subscription is cancelled or not'),
    )

    class Meta:
        ordering = ('user', 'date_billing_start',)
        abstract = True

    def record_transaction(self, amount=None, transaction_date=None):
        """Records transaction details in SubscriptionTransaction.
            Parameters:
                amount: Use custom amount to create transaction for the subscription
                transaction_date (obj): A DateTime object of when
                    payment occurred (defaults to current datetime if
                    none provided).
            Returns:
                obj: The created SubscriptionTransaction instance.
        """
        if transaction_date is None:
            transaction_date = timezone.now()

        if amount is None:
            amount = self.plan_cost.cost
        SubscriptionTransaction = swapper.load_model('subscriptions_api', 'SubscriptionTransaction')
        return SubscriptionTransaction.objects.create(
            user=self.user,
            subscription=self,  # A transaction should link to is subscription
            date_transaction=transaction_date,
            amount=amount,
        )

    @property
    def unused_daily_balance(self):
        """
        Calculate unused balance  of a subscription for a particular period
        """
        current_date = timezone.now()
        if self.date_billing_next > current_date:
            days_left = (self.date_billing_next - current_date).days
            return round(days_left * self.plan_cost.daily_cost, 2)
        return 0

    @property
    def description(self):
        if self.plan_cost:
            text = f'{self.plan_cost.plan.plan_name} {self.plan_cost.display_billing_frequency_text}'
            return text

    @property
    def used_daily_balance(self):
        """
        Calculate used balance  of a subscription for a particular period
        """
        current_date = timezone.now()
        if current_date > self.date_billing_start:
            days_used = (current_date - self.date_billing_start).days
            return round(days_used * self.plan_cost.daily_cost, 2)
        return 0

    def activate(self, subscription_date=None, mark_transaction_paid=True):
        current_date = subscription_date or timezone.now()
        next_billing_date = self.plan_cost.next_billing_datetime(current_date)
        self.active = True
        self.cancelled = False
        self.due = False
        self.date_billing_start = current_date
        self.date_billing_end = next_billing_date + timedelta(days=self.plan_cost.plan.grace_period)
        self.date_billing_next = next_billing_date
        self._add_user_to_group()
        if mark_transaction_paid:
            self.transactions.update(paid=True)
        self.save()

    def deactivate(self):
        current_date = timezone.now()
        self.active = False
        self.date_billing_last = current_date
        self.cancelled = True
        self.due = False
        self._remove_user_from_group()
        self.save()

    def _add_user_to_group(self):
        try:
            group = self.plan_cost.plan.group
            group.user_set.add(self.user)
        except AttributeError:
            # No group available to add user to
            pass

    def _remove_user_from_group(self):
        try:
            group = self.plan_cost.plan.group
            group.user_set.remove(self.user)
        except AttributeError:
            # No group available to add user to
            pass

    def notify(self, notifier, **kwargs):
        """
        param notififer: notifie class that takes usersubscription object and has a send method
        """
        if SETTINGS[notifier] is None:
            return
        Notify = getattr(  # pylint: disable=invalid-name
            importlib.import_module(SETTINGS[notifier]['module']),
            SETTINGS[notifier]['class']
        )
        notify_obj = Notify(self, notifier, **kwargs)
        notify_obj.send()
        return notify_obj

    def notify_processing(self, **kwargs):
        """Sends notification of processing subscription.

        """
        return self.notify('notify_processing', **kwargs)

    def notify_expired(self, **kwargs):
        """Sends notification of expired subscription.

        """
        return self.notify('notify_expired', **kwargs)

    def notify_overdue(self, **kwargs):
        """Sends notification of overdue subscription.

        """
        return self.notify('notify_overdue', **kwargs)

    def notify_new(self, **kwargs):
        """Sends notification of new subscription

        """
        return self.notify('notify_new', **kwargs)

    def notify_activate(self, **kwargs):
        """Sends notification of activated subscription

        """
        return self.notify('notify_activate', **kwargs)

    def notify_deactivate(self, **kwargs):
        """Sends notification of deactivated subscription

        """
        return self.notify('notify_deactivate', **kwargs)

    def notify_payment_error(self, **kwargs):
        """Sends notification of a payment error

        """
        return self.notify('notify_payment_error', **kwargs)

    def notify_payment_success(self, **kwargs):
        """Sends notifiation of a payment success

        """
        return self.notify('notify_payment_success', **kwargs)


class BaseSubscriptionTransaction(models.Model):
    """Details for a subscription plan billing."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    user = models.ForeignKey(
        get_user_model(),
        help_text=_('the user that this subscription was billed for'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='subscription_transactions'
    )
    subscription = models.ForeignKey(
        UserSubscriptionModel,
        help_text=_('the Subscription that were billed'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='transactions'
    )
    date_transaction = models.DateTimeField(
        help_text=_('the datetime the transaction was billed'),
        verbose_name='transaction date',
    )
    amount = models.DecimalField(
        blank=True,
        decimal_places=2,
        help_text=_('how much was billed for the user'),
        max_digits=19,
        null=True,
    )

    paid = models.BooleanField(default=False, help_text=_('Mark transaction has paid'))

    class Meta:
        ordering = ('date_transaction', 'user',)
        abstract = True



# Generated by Django 3.0.5 on 2020-05-12 17:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions_api', '0002_auto_20200512_1405'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscriptionplan',
            name='feature_ref',
            field=models.CharField(blank=True, help_text='Reference to select list of allowed features for this plan', max_length=100, null=True, unique=True),
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-12 14:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions_api', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='planlistdetail',
            options={'ordering': ('order',)},
        ),
        migrations.AddField(
            model_name='planlist',
            name='features_ref',
            field=models.CharField(blank=True, help_text='Reference to select list of features to display for the plan list', max_length=12, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='planlistdetail',
            name='html_content',
            field=models.TextField(blank=True, help_text='HTML content can also be json object to display for plan', null=True),
        ),
    ]



import inspect
import importlib
import json
from subscriptions_api.app_settings import SETTINGS
from subscriptions_api.models import SubscriptionPlan


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def get_plans_module():
    module_str = SETTINGS['plans_concrete_module']
    module = importlib.import_module(module_str)
    return inspect.getmembers(module)


def get_plans():
    plans_data = {}
    for name, obj in get_plans_module():
        if isinstance(obj, dict) and name.endswith('_PLAN'):
            obj = dotdict(obj)
            plans_data[obj._name] = obj
    return plans_data


def get_plan_feature(plan_name):
    if not plan_name:
        return None
    # Get features from model
    plan = SubscriptionPlan.objects.get(plan_name=plan_name)
    try:
        # If features of a plan is stored in json then send it
        features_dict = json.loads(plan.features)
        return dotdict(features_dict)
    except json.JSONDecodeError:
        return get_plans().get(plan_name, None)


def get_planlist_features(feature_ref=None):
    for name, obj in get_plans_module():
        if isinstance(obj, dict):
            if feature_ref and name == 'FEATURES_%s' % (feature_ref):
                return obj
            if not feature_ref and name == 'DEFAULT_FEATURES':
                return obj
    return {}


BASE_PLAN = {'_name': 'BasePlan'}
FEATURES_ONE = {'_name': 'Plan Name'}



"""Models Gotten from the Flexible Subscriptions app.
    with minor changes
"""
from datetime import timedelta
from uuid import uuid4
import swapper
import json
from django.contrib.auth.models import Group
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from subscriptions_api.base_models import BaseUserSubscription, BaseSubscriptionTransaction

# Convenience references for units for plan recurrence billing
# ----------------------------------------------------------------------------
ONCE = '0'
SECOND = '1'
MINUTE = '2'
HOUR = '3'
DAY = '4'
WEEK = '5'
MONTH = '6'
YEAR = '7'
RECURRENCE_UNIT_CHOICES = (
    (ONCE, 'once'),
    (SECOND, 'second'),
    (MINUTE, 'minute'),
    (HOUR, 'hour'),
    (DAY, 'day'),
    (WEEK, 'week'),
    (MONTH, 'month'),
    (YEAR, 'year'),
)


# ----------------------------------------------------------------------------

class UserSubscription(BaseUserSubscription):
    class Meta:
        swappable = swapper.swappable_setting('subscriptions_api', 'UserSubscription')


class SubscriptionTransaction(BaseSubscriptionTransaction):
    class Meta:
        swappable = swapper.swappable_setting('subscriptions_api', 'SubscriptionTransaction')


class PlanTag(models.Model):
    """A tag for a subscription plan."""
    tag = models.CharField(
        help_text=_('the tag name'),
        max_length=64,
        unique=True,
    )

    class Meta:
        ordering = ('tag',)

    def __str__(self):
        return self.tag


class SubscriptionPlan(models.Model):
    """Details for a subscription plan."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    plan_name = models.CharField(
        help_text=_('the name of the subscription plan'),
        max_length=128,
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference the subscription plan'),
        max_length=128,
        null=True,
        unique=True,
    )
    feature_ref = models.CharField(
        blank=True,
        help_text=_('Reference to select list of allowed features for this plan'),
        max_length=100,
        null=True,
        unique=True,
    )
    features = models.TextField(
        blank=True,
        help_text=_('list of json allowed features for this plan'),
        null=True
    )
    plan_description = models.CharField(
        blank=True,
        help_text=_('a description of the subscription plan'),
        max_length=512,
        null=True,
    )
    group = models.ForeignKey(
        Group,
        blank=True,
        help_text=_('the Django auth group for this plan'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='plans',
    )
    tags = models.ManyToManyField(
        PlanTag,
        blank=True,
        help_text=_('any tags associated with this plan'),
        related_name='plans',
    )
    grace_period = models.PositiveIntegerField(
        default=0,
        help_text=_(
            'how many days after the subscription ends before the '
            'subscription expires'
        ),
    )
    trial_period = models.PositiveIntegerField(
        default=0,
        help_text=_(
            'how many days to give free before the subscription '
            'subscription begins'
        ),
    )

    class Meta:
        ordering = ('plan_name',)
        permissions = (
            ('subscriptions', 'Can interact with subscription details'),
        )

    def __getattr__(self, name):
        if self.features:
            feature_dict = json.loads(self.features)
            try:
                return feature_dict[name]
            except KeyError:
                pass
        raise AttributeError

    def __str__(self):
        return self.plan_name

    def display_tags(self):
        """Displays tags as a string (truncates if more than 3)."""
        if self.tags.count() > 3:
            return '{}, ...'.format(
                ', '.join(tag.tag for tag in self.tags.all()[:3])
            )

        return ', '.join(tag.tag for tag in self.tags.all()[:3])


class PlanCost(models.Model):
    """Cost and frequency of billing for a plan."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    plan = models.ForeignKey(
        SubscriptionPlan,
        help_text=_('the subscription plan for these cost details'),
        on_delete=models.CASCADE,
        related_name='costs',
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference these cost details'),
        max_length=128,
        null=True,
        unique=True,
    )
    recurrence_period = models.PositiveSmallIntegerField(
        default=1,
        help_text=_('how often the plan is billed (per recurrence unit)'),
        validators=[MinValueValidator(1)],
    )
    recurrence_unit = models.CharField(
        choices=RECURRENCE_UNIT_CHOICES,
        default=MONTH,
        max_length=1,
    )
    cost = models.DecimalField(
        blank=True,
        decimal_places=2,
        help_text=_('the cost per recurrence of the plan'),
        max_digits=19,
        null=True,
    )

    class Meta:
        ordering = ('recurrence_unit', 'recurrence_period', 'cost',)

    @property
    def display_recurrent_unit_text(self):
        """Converts recurrence_unit integer to text."""
        conversion = {
            ONCE: 'one-time',
            SECOND: 'per second',
            MINUTE: 'per minute',
            HOUR: 'per hour',
            DAY: 'per day',
            WEEK: 'per week',
            MONTH: 'per month',
            YEAR: 'per year',
        }

        return conversion[self.recurrence_unit]

    @property
    def display_billing_frequency_text(self):
        """Generates human-readable billing frequency."""
        conversion = {
            ONCE: 'one-time',
            SECOND: {'singular': 'per second', 'plural': 'seconds'},
            MINUTE: {'singular': 'per minute', 'plural': 'minutes'},
            HOUR: {'singular': 'per hour', 'plural': 'hours'},
            DAY: {'singular': 'per day', 'plural': 'days'},
            WEEK: {'singular': 'per week', 'plural': 'weeks'},
            MONTH: {'singular': 'per month', 'plural': 'months'},
            YEAR: {'singular': 'per year', 'plural': 'years'},
        }

        if self.recurrence_unit == ONCE:
            return conversion[ONCE]

        if self.recurrence_period == 1:
            return conversion[self.recurrence_unit]['singular']

        return 'every {} {}'.format(
            self.recurrence_period, conversion[self.recurrence_unit]['plural']
        )

    def next_billing_datetime(self, current):
        """Calculates next billing date for provided datetime.
            Parameters:
                current (datetime): The current datetime to compare
                    against.
            Returns:
                datetime: The next time billing will be due.
        """
        if self.recurrence_unit == SECOND:
            return current + timedelta(seconds=self.recurrence_period)

        if self.recurrence_unit == MINUTE:
            return current + timedelta(minutes=self.recurrence_period)

        if self.recurrence_unit == HOUR:
            return current + timedelta(hours=self.recurrence_period)

        if self.recurrence_unit == DAY:
            return current + timedelta(days=self.recurrence_period)

        if self.recurrence_unit == WEEK:
            return current + timedelta(weeks=self.recurrence_period)

        if self.recurrence_unit == MONTH:
            # Adds the average number of days per month as per:
            # http://en.wikipedia.org/wiki/Month#Julian_and_Gregorian_calendars
            # This handle any issues with months < 31 days and leap years
            return current + timedelta(
                days=30.4368 * self.recurrence_period
            )

        if self.recurrence_unit == YEAR:
            # Adds the average number of days per year as per:
            # http://en.wikipedia.org/wiki/Year#Calendar_year
            # This handle any issues with leap years
            return current + timedelta(
                days=365.2425 * self.recurrence_period
            )

        return None

    def setup_user_subscription(self, user, active=True, subscription_date=None, no_multipe_subscription=False,
                                del_multipe_subscription=False, record_transaction=False, mark_transaction_paid=True,
                                resuse=False):
        """Adds subscription to user and adds them to required group if active.
            Parameters:
                user (obj): A Django user instance.
                active (bool): Add user to required group if active.
                subscription_date (date) :  Date to use for  creation subscription
            Returns:
                obj: The newly created UserSubscription instance.
        """
        if no_multipe_subscription:
            previous_subscriptions = user.subscriptions.filter(active=True).all()
            for sub in previous_subscriptions:
                sub.deactivate()
                if del_multipe_subscription:
                    sub.delete()

        # Add subscription plan to user
        subscription = None
        if resuse:
            subscription = user.subscriptions.filter(plan_cost=self).first()
        if not subscription:
            UserSubscription = swapper.load_model('subscriptions_api', 'UserSubscription')

            subscription = UserSubscription.objects.create(
                user=user,
                plan_cost=self,
                active=active,
                cancelled=False
            )
        # Add user to the proper group
        if record_transaction:
            subscription.record_transaction(transaction_date=subscription_date)
        if active:
            subscription.activate(subscription_date=subscription_date, mark_transaction_paid=mark_transaction_paid)
        return subscription

    @property
    def daily_cost(self):
        """
        Returns the daily cost for a plan cost period
        """
        if self.recurrence_unit == DAY:
            return float(self.cost) / (1 * self.recurrence_period)

        if self.recurrence_unit == WEEK:
            return float(self.cost) / (7 * self.recurrence_period)

        if self.recurrence_unit == MONTH:
            return float(self.cost) / (30.4368 * self.recurrence_period)

        if self.recurrence_unit == YEAR:
            # Adds the average number of days per year as per:
            # http://en.wikipedia.org/wiki/Year#Calendar_year
            # This handle any issues with leap years
            return float(self.cost) / (365.2425 * self.recurrence_period)

        return 0

    def __str__(self):
        return '{} {} {}'.format(self.plan.plan_name, self.display_billing_frequency_text, self.cost)


class PlanList(models.Model):
    """Model to record details of a display list of SubscriptionPlans."""
    title = models.TextField(
        blank=True,
        help_text=_('title to display on the subscription plan list page'),
        null=True,
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference the subscription plan list'),
        max_length=128,
        null=True,
        unique=True,
    )
    features_ref = models.CharField(
        blank=True,
        help_text=_('Reference to select list of features to display for the plan list'),
        max_length=12,
        null=True,
        unique=True,
    )
    subtitle = models.TextField(
        blank=True,
        help_text=_('subtitle to display on the subscription plan list page'),
        null=True,
    )
    header = models.TextField(
        blank=True,
        help_text=_('header text to display on the subscription plan list page'),
        null=True,
    )
    footer = models.TextField(
        blank=True,
        help_text=_('header text to display on the subscription plan list page'),
        null=True,
    )
    active = models.BooleanField(
        default=True,
        help_text=_('whether this plan list is active or not.'),
    )

    def __str__(self):
        return self.title


class PlanListDetail(models.Model):
    """Model to add additional details to plans when part of PlanList."""
    plan = models.ForeignKey(
        SubscriptionPlan,
        on_delete=models.CASCADE,
        related_name='plan_list_details',
    )
    plan_list = models.ForeignKey(
        PlanList,
        on_delete=models.CASCADE,
        related_name='plan_list_details',
    )
    html_content = models.TextField(
        blank=True,
        help_text=_('HTML content can also be json object to display for plan'),
        null=True,
    )
    subscribe_button_text = models.CharField(
        blank=True,
        default='Subscribe',
        max_length=128,
        null=True,
    )
    order = models.PositiveIntegerField(
        default=1,
        help_text=_('Order to display plan in (lower numbers displayed first)'),
    )

    def __str__(self):
        return 'Plan List {} - {}'.format(
            self.plan_list, self.plan.plan_name
        )

    class Meta:
        ordering = ('order',)



from django.conf import settings


def string_to_module_and_class(string):
    """Breaks a string to a module and class name component."""
    components = string.split('.')
    component_class = components.pop()
    component_module = '.'.join(components)

    return {
        'module': component_module,
        'class': component_class,
    }


def compile_settings():
    """Compiles and validates all package settings and defaults.
        Provides basic checks to ensure required settings are declared
        and applies defaults for all missing settings.
        Returns:
            dict: All possible Django Flexible Subscriptions settings.
    """
    plans_concrete_module = getattr(
        settings, 'DFS_CONCRETE_PLANS_MODULE', 'subscriptions_api.plans'
    )
    subscribe_notify_processing = getattr(
        settings, 'DFS_NOTIFY_PROCESSING', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_processing_class = string_to_module_and_class(subscribe_notify_processing)

    subscribe_notify_expired = getattr(
        settings, 'DFS_NOTIFY_EXPIRED', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_expired_class = string_to_module_and_class(subscribe_notify_expired)

    subscribe_notify_overdue = getattr(
        settings, 'DFS_NOTIFY_OVERDUE', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_overdue_class = string_to_module_and_class(subscribe_notify_overdue)

    subscribe_notify_new = getattr(
        settings, 'DFS_NOTIFY_NEW', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_new_class = string_to_module_and_class(subscribe_notify_new)

    subscribe_notify_payment_error = getattr(
        settings, 'DFS_NOTIFY_ERROR', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_payment_error_class = string_to_module_and_class(subscribe_notify_payment_error)

    subscribe_notify_payment_success = getattr(
        settings, 'DFS_NOTIFY_PAYMENT_SUCCESS', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_payment_success_class = string_to_module_and_class(subscribe_notify_payment_success)

    subscribe_notify_activate = getattr(
        settings, 'DFS_NOTIFY_ACTIVATE', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_activate_class = string_to_module_and_class(subscribe_notify_activate)

    subscribe_notify_deactivate = getattr(
        settings, 'DFS_NOTIFY_DEACTIVATE', 'subscriptions_api.notifications.EmailNotification'
    )
    subscribe_notify_deactivate_class = string_to_module_and_class(subscribe_notify_deactivate)

    return {
        'notify_processing': subscribe_notify_processing_class,
        'notify_expired': subscribe_notify_expired_class,
        'notify_overdue': subscribe_notify_overdue_class,
        'notify_new': subscribe_notify_new_class,
        'notify_activate': subscribe_notify_activate_class,
        'notify_deactivate': subscribe_notify_deactivate_class,
        'notify_payment_error': subscribe_notify_payment_error_class,
        'notify_payment_success': subscribe_notify_payment_success_class,
        'plans_concrete_module': plans_concrete_module
    }


SETTINGS = compile_settings()
